/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id$
 *
 * Copyright (c) 2000 Tilo Christ. All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **********************************************************************************/

#ifndef CANVAS_H
#define CANVAS_H

#include <PalmOS.h>


/**
 * Canvas manages a canvas into which a user can paint.
 */
class Canvas
{
	public:

///@name Construction / Destruction
//@{
	/**
	 * Construct a new canvas.
	 */
	Canvas()
	{
		WinGetDisplayExtent(&width, &height);
		
		displayBounds.topLeft.x = 0;
		displayBounds.topLeft.y = 0;
		displayBounds.extent.x = width;
		displayBounds.extent.y = height;
	}


	/**
	 * Destroy the canvas.
	 */
	virtual ~Canvas() {}
//@}


/// @name Output management
//@{
	/**
	 * Switch drawing window to the window into which the application shall paint.
	 */
	virtual void beginDraw(WinLockInitType initMode) = 0;
	
	/**
	 * Notify the canvas of the end of user paint operations.
	 * 
	 * @param bounds the area that was modified by the user draw operation
	 */
	virtual void endDraw(RectangleType *bounds) = 0;

	/**
	 * Display user painted graphics.
	 */
	virtual void show() = 0;	
//@}


/// @name Physical Display management
//@{
	/**
	 * Get the width of the display
	 */
	Coord getWidth() const
	{
		return width;
	}

	
	/**
	 * Get the height of the display
	 */
	Coord getHeight() const
	{
		return height;
	}
//@}



/// @name Utility operations
//@{
	/**
	 * Determine a rectangle that encompasses two other rectangles.
	 *
	 * @param rect1Bound the bounds of rectangle #1
	 * @param rect2Bound the bounds of rectangle #2
	 * @param resultBound the bounds of the encompassing rectangle.
	          rect1Bound, or rect2Bound, and resultBound may safely point to the same RectangleType structure.
	 */
	static void uniteBounds(RectangleType *rect1Bound, RectangleType *rect2Bound, RectangleType *resultBound)
	{
		Coord rect1RightEdge = rect1Bound->topLeft.x + rect1Bound->extent.x;
		Coord rect2RightEdge = rect2Bound->topLeft.x + rect2Bound->extent.x;

		if (rect1Bound->topLeft.x < rect2Bound->topLeft.x)
			resultBound->topLeft.x = rect1Bound->topLeft.x;
		else
			resultBound->topLeft.x = rect2Bound->topLeft.x;

		if (rect1RightEdge > rect2RightEdge)
			resultBound->extent.x = rect1RightEdge - resultBound->topLeft.x;
		else
			resultBound->extent.x = rect2RightEdge - resultBound->topLeft.x;


		Coord rect1BottomEdge = rect1Bound->topLeft.y + rect1Bound->extent.y;
		Coord rect2BottomEdge = rect2Bound->topLeft.y + rect2Bound->extent.y;

		if (rect1Bound->topLeft.y < rect2Bound->topLeft.y)
			resultBound->topLeft.y = rect1Bound->topLeft.y;
		else
			resultBound->topLeft.y = rect2Bound->topLeft.y;

		if (rect1BottomEdge > rect2BottomEdge)
			resultBound->extent.y = rect1BottomEdge - resultBound->topLeft.y;
		else
			resultBound->extent.y = rect2BottomEdge - resultBound->topLeft.y;
	}
//@}

	protected:
	RectangleType displayBounds;

	private:
	Coord width;
	Coord height;
};


#endif
