/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id$
 *
 * Copyright (c) 2000 Tilo Christ. All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **********************************************************************************/



#ifndef SPRITE_ENGINE_H
#define SPRITE_ENGINE_H

#include <PalmOS.h>
#include "Canvas.h"
#include "SoundEngine.h"        // Needed for invocations of SoundEngine::timeTick()


class SpriteGroup;


/**
 * Sprite describes a single sprite, i.e. a movable graphical object
 * with a transparent background.
 */
class Sprite
{
	public:

	enum OptimizationMode
	{
		optNONE       = 0,
		optBUFFER     = 1,
		optNO_MASK    = 2
	};

	
	/**
	 * Create a Sprite
	 *
	 * @param bitmapID the ID of the image bitmap in the resource database.
	 * @param maskID the ID of the mask bitmap in the resource database.
	 * @param hotSpotX the horizontal offset of the hotspot.
	 * @param hotSpotY the vertical offset of the hotspot.
	 * @param visible shall the sprite be visible?
	 * @param optimizationMode which optimizations shall be applied?
	 */
	Sprite(DmResID bitmapID, DmResID maskID = -1, Coord hotSpotX = 0, Coord hotSpotY = 0, Boolean visible = true, OptimizationMode optimizationMode = optNONE)
	{
		this->bitmapID = bitmapID;
		this->maskID = maskID;
		this->hotSpotX = hotSpotX;
		this->hotSpotY = hotSpotY;		
		this->visible = visible;
		this->optimizationMode = optimizationMode;

		move(0,0);

		switch(optimizationMode)
		{
		case optBUFFER:
			{
			WinHandle oldDrawH = WinGetDrawWindow();
			UInt16 error = errNone;

			{
				// Paint sprite-bitmap to offscreen window
				MemHandle resH = DmGetResource( bitmapRsc, bitmapID );
				ErrNonFatalDisplayIf( !resH, "Missing bitmap" );
				BitmapType* resP = (BitmapPtr)MemHandleLock(resH);

				this->width = resP->width;
				this->height = resP->height;

				offScreenBitmap = WinCreateOffscreenWindow(resP->width, resP->height, screenFormat, &error);
				WinSetDrawWindow(offScreenBitmap);
				WinDrawBitmap (resP, 0,0);

				MemPtrUnlock(resP);
				DmReleaseResource( resH );
			}

			{
				// Paint sprite-mask to offscreen window
				MemHandle resH = DmGetResource( bitmapRsc, maskID );
				ErrNonFatalDisplayIf( !resH, "Missing mask" );
				BitmapType* resP = (BitmapPtr)MemHandleLock(resH);

				offScreenMask = WinCreateOffscreenWindow(resP->width, resP->height, screenFormat, &error);
				WinSetDrawWindow(offScreenMask);
				WinDrawBitmap (resP, 0,0);
				MemPtrUnlock(resP);
				DmReleaseResource( resH );				
			}

			WinSetDrawWindow(oldDrawH);
			}			
			break;
		
		case optNONE:
		case optNO_MASK:
			// Lookup Bitmap resource
			MemHandle resH = DmGetResource( bitmapRsc, bitmapID );
			ErrNonFatalDisplayIf( !resH, "Missing bitmap" );
			BitmapType* resP = (BitmapPtr)MemHandleLock(resH);

			this->width = resP->width;
			this->height = resP->height;

			MemPtrUnlock(resP);
			DmReleaseResource( resH );

			offScreenBitmap = NULL;
			offScreenMask   = NULL;
			
			break;
		}
	}
	
	
	/**
	 * Destroy the Sprite and deallocate all used resources.
	 */
	~Sprite()
	{
		if (offScreenBitmap != NULL)
		{
			WinDeleteWindow(offScreenBitmap, false);
		}

		if (offScreenMask != NULL)
		{
			WinDeleteWindow(offScreenMask, false);
		}
	}


	/**
	 * Show the sprite during subsequent draws.
	 */
	void show()
	{
		visible = true;
	}


	/**
	 * Hide the sprite during subsequent draws.
	 */
	void hide()
	{
		visible = false;
	}


	/**
	 * Set the visibility of the sprite through a flag.
     *
	 * @param visible shall the sprite be visible (true = shown / false = hidden)
	 */
	Boolean setVisibility(Boolean visible)
	{
		Boolean oldVisible = visible;
		this->visible = visible;
		
		return oldVisible;
	}


	/**
	 * Will the sprite be drawn during subsequent draws?
	 */
	Boolean isVisible() const
	{
		return visible;
	}


	/**
	 * Move the sprite (i.e. its hotspot) to the specified coordinates.
	 */
	void move(Coord x, Coord y)
	{
		this->x = x;
		this->y = y;
	}


	/**
	 * Draw the sprite at its current location.
	 * Inquire the current draw window and allocate draw buffers as neccessary.
     *
     * @param bounds a pointer to a rectangle which will be filled with the screen space
     *        filled by the sprite, or NULL.
	 */
	void draw(RectangleType *bounds = NULL) const
	{
		if (!visible)
		{
			if (bounds != NULL)
			{
				bounds->topLeft.x = 0;
				bounds->topLeft.y = 0;
				bounds->extent.x = 0;
				bounds->extent.y = 0;
			}
			
			return;
		}
		
	
		UInt16 error;
		
		WinHandle drawWindow = WinGetDrawWindow();

		switch(optimizationMode)
		{
		case optBUFFER:
		case optNO_MASK:
			quickDraw(drawWindow, NULL, bounds);
			
			break;

		case optNONE:
			WinHandle offScreenH = WinCreateOffscreenWindow(width, height, screenFormat, &error);
			WinSetDrawWindow(offScreenH);
			quickDraw(drawWindow, offScreenH, bounds);
			WinSetDrawWindow(drawWindow);
			WinDeleteWindow(offScreenH, false);

			break;
		}
	}


	/**
	 * Get the screen space filled by the sprite.
	 *
     * @param bounds a pointer to a rectangle which will be filled with the screen space
     *        filled by the sprite.
	 */
	void getBounds(RectangleType *bounds) const
	{
		ErrNonFatalDisplayIf(bounds == NULL, "getBounds: bounds == NULL");
	
		bounds->topLeft.x = x - hotSpotX;
		bounds->topLeft.y = y - hotSpotY;
		bounds->extent.x = width;
		bounds->extent.y = height;
	}



	private:

	
	/**
	 * Draw the sprite at its current location into the specified drawWindow 
	 * and use the specified offScreen buffer for internal drawing operations.
	 */
	void quickDraw(WinHandle drawWindow, WinHandle offScreenH, RectangleType *bounds) const
	{
		if (!visible)
		{
			if (bounds != NULL)
			{
				bounds->topLeft.x = 0;
				bounds->topLeft.y = 0;
				bounds->extent.x = 0;
				bounds->extent.y = 0;
			}
			
			return;
		}
		else
		{
			if (bounds != NULL)
				getBounds(bounds);
		}	

		RectangleType size;
		size.topLeft.x = 0;
		size.topLeft.y = 0;
		size.extent.x = width;
		size.extent.y = height;			
		
		
		switch(optimizationMode)
		{
		case optBUFFER:
			// The fast drawing code...		
		
			WinCopyRectangle(offScreenMask, drawWindow, &size, x - hotSpotX, y - hotSpotY, winMask);
			SoundEngine::timeTick();
			WinCopyRectangle(offScreenBitmap, drawWindow, &size, x - hotSpotX, y - hotSpotY, winOverlay);
			SoundEngine::timeTick();
		
			break;	

		case optNONE:
			// The slow drawing code...
			{
				// Paint sprite-mask to offscreen window
				MemHandle resH = DmGetResource( bitmapRsc, maskID );
				ErrNonFatalDisplayIf( !resH, "Missing mask" );
				BitmapType* resP = (BitmapPtr)MemHandleLock(resH);

				WinDrawBitmap (resP, 0,0);

				MemPtrUnlock(resP);
				DmReleaseResource( resH );				

				WinCopyRectangle(offScreenH, drawWindow, &size, x - hotSpotX, y - hotSpotY, winMask);
	
				SoundEngine::timeTick();
			}
			
			{
				// Paint sprite-overlay over the masked-out area
				MemHandle resH = DmGetResource( bitmapRsc, bitmapID );
				ErrNonFatalDisplayIf( !resH, "Missing bitmap" );
				BitmapType* resP = (BitmapPtr)MemHandleLock(resH);

				WinDrawBitmap (resP, 0,0);

				MemPtrUnlock(resP);
				DmReleaseResource( resH );

				WinCopyRectangle(offScreenH, drawWindow, &size, x - hotSpotX, y - hotSpotY, winOverlay);

				SoundEngine::timeTick();	
			}
		
			break;
			
			
		case optNO_MASK:
			{
				MemHandle resH = DmGetResource( bitmapRsc, bitmapID );
				ErrNonFatalDisplayIf( !resH, "Missing bitmap" );
				BitmapType* resP = (BitmapPtr)MemHandleLock(resH);

				WinDrawBitmap (resP, x - hotSpotX, y - hotSpotY);

				MemPtrUnlock(resP);
				DmReleaseResource( resH );

				SoundEngine::timeTick();	
			}	
		}
	}
	

	OptimizationMode getOptimizationMode() const
	{
		return optimizationMode;
	}



	DmResID bitmapID;
	DmResID maskID;
	Coord hotSpotX;
	Coord hotSpotY;
	Int16 width;
	Int16 height;
	Boolean visible;
	OptimizationMode optimizationMode;

	WinHandle offScreenBitmap;
	WinHandle offScreenMask;
	
	Coord x;
	Coord y;


	// Ain't it nice to have friends...
	friend class SpriteGroup;
};



/**
 * SpriteGroup assists in the management of a collection of sprites.
 */
class SpriteGroup
{
	public:
	
	
	/**
	 * Create the SpriteGroup. The group may contain up to maxSprites Sprites.
	 * @param maxSprites the maximum number of managed sprites
	 */
	SpriteGroup(int maxSprites)
	{
		ErrNonFatalDisplayIf(maxSprites == 0, "maxSprites == 0");
		this->maxSprites = maxSprites;
		this->currSprites = 0;
		
		this->theSprites = new Sprite*[maxSprites];

		this->offscreenH = NULL;
		this->maxWidth = 0;
		this->maxHeight = 0;
		this->oldMaxWidth = -1;
		this->oldMaxHeight = -1;
		this->needBuffer = false;
	}
	
	
	/**
	 * Destroy the SpriteGroup and deallocate all used resources.
	 */
	~SpriteGroup()
	{
		delete [] theSprites;
		
		if (offscreenH != NULL)
			WinDeleteWindow(offscreenH, false);
	}


	/**
	 * Add a sprite to the group.
	 *
	 * @param sprite a reference to the sprite. Ownership of the sprite will remain
	 *        with the caller, i.e. SpriteGroup will not destroy the sprite when the
	 *        group is destroyed.
	 */
	void addSprite(Sprite& sprite)
	{
		ErrNonFatalDisplayIf(currSprites == maxSprites, "SpriteGroup size exhausted");

		this->theSprites[currSprites] = &sprite;

		// Adjust size of drawing buffer if neccessary
		if (sprite.getOptimizationMode() == Sprite::optNONE)
		{
			RectangleType bounds;
			sprite.getBounds(&bounds);

			maxWidth  = max(bounds.extent.x, maxWidth);
			maxHeight = max(bounds.extent.y, maxHeight);
			
			needBuffer = true;
		}

		currSprites++;
	}


	/**
	 * Draw all the sprites in the group.
	 */
	void draw(RectangleType *bounds)
	{
		if (currSprites == 0)
			return;

		WinHandle drawWindow = WinGetDrawWindow();
		if (needBuffer)
		{
			if ((maxWidth != oldMaxWidth) || (maxHeight != oldMaxHeight))
			{
				if (offscreenH != NULL)
				{
					WinDeleteWindow(offscreenH, false);
					offscreenH = NULL;
				}
			
				oldMaxWidth = maxWidth;
				oldMaxHeight = maxHeight;
			
				UInt16 error;
				offscreenH = WinCreateOffscreenWindow(maxWidth, maxHeight, screenFormat, &error);
			}

			WinSetDrawWindow(offscreenH);
		}
	

		RectangleType spriteBounds;
		for(int i = 0; i < currSprites; i++)
		{
			this->theSprites[i]->quickDraw(drawWindow, offscreenH, &spriteBounds);
			Canvas::uniteBounds(&spriteBounds, bounds, bounds);
		}

		if (needBuffer)
			WinSetDrawWindow(drawWindow);
	}


	private:

	int maxSprites;
	int currSprites;
	Sprite* *theSprites;
	
	WinHandle offscreenH;
	int maxWidth;
	int maxHeight;
	int oldMaxWidth;
	int oldMaxHeight;
	Boolean needBuffer;
};



#endif
